@@ -1,4 +1,7 @@
-from __future__ import absolute_import
+# Copyright (C) The Arvados Authors. All rights reserved.
+#
+# SPDX-License-Identifier: Apache-2.0
+
 import traceback
 import unittest
 
